﻿unit uActorNodalCombustible;

{$IFDEF FPC}
  {$MODE Delphi}
{$ENDIF}

interface

uses
  uCosa, uGlobs, uActores, uFechas, unodoCombustible, uCosaConNombre, uFichasLPD,
  xMatDefs, uauxiliares;

type
  TActorNodalCombustible = class(TActor)
    function ReferenciaAlNodoCombustible (nodocomb: TNodoCombustible): boolean; virtual; abstract;
  end;

  //TActorUniNodalCombustible es un nodo NO ELÉCTRICO

  TActorUniNodalCombustible = class(TActorNodalCombustible)
  public
    nodocomb: TNodoCombustible;
    Q: TDAOfNReal;   // caudal inyectado al nodo

    constructor Create(capa: integer; nombre: string; nacimiento, muerte: TFecha;
      lpdUnidades: TFichasLPD; nodocomb: TNodoCombustible);
    constructor Create_ReadFromText(f: TArchiTexto); override;
    procedure WriteToText(f: TArchiTexto); override;
    function ReferenciaAlNodoCombustible (nodocomb: TNodoCombustible): boolean; override;
    procedure PubliVars; override;
    procedure PrepararMemoria(globs: TGlobs); override;
    procedure Free; override;
  end;

  TActorBiNodalCombustible = class(TActorNodalCombustible)
  public
    NodoCombA, NodoCombB: TNodoCombustible;
    Q_NodoCombA, Q_NodoCombB: TDAOfNReal;

    constructor Create(capa: integer; nombre: string; nacimiento, muerte: TFecha;
      lpdUnidades: TFichasLPD; NodoCombA, NodoCombB: TNodoCombustible);
    constructor Create_ReadFromText(f: TArchiTexto); override;
    procedure WriteToText(f: TArchiTexto); override;
    function ReferenciaAlNodoCombustible (nodocomb: TNodoCombustible): boolean; override;
    procedure PrepararMemoria(globs: TGlobs); override;
    procedure Free; override;
  end;

procedure AlInicio;
procedure ALFinal;

implementation

//--------------------------
// Métodos de TActorUniNodalCombustible
//==========================

constructor TActorUniNodalCombustible.Create(capa: integer; nombre: string; nacimiento, muerte: TFecha;
  lpdUnidades: TFichasLPD; NodoComb: TNodoCombustible);
begin
  inherited Create(capa, nombre, nacimiento, muerte, lpdUnidades);
  self.NodoComb := NodoComb;
end;

constructor TActorUniNodalCombustible.Create_ReadFromText(f: TArchiTexto);
begin
  inherited Create_ReadFromText(f);
  f.IniciarLecturaRetrasada;
  f.rdReferencia('nodo combustible', TCosa(nodocomb), Self);
  f.EjecutarLectura;

end;

procedure TActorUniNodalCombustible.WriteToText(f: TArchiTexto);
begin
  inherited WriteToText(f);
  f.wrReferencia('nodo combustible', nodocomb);
end;

function TActorUniNodalCombustible.referenciaAlNodoCombustible (nodocomb: TNodoCombustible): boolean;
begin
  Result := self.nodocomb = nodocomb;
end;

procedure TActorUniNodalCombustible.PubliVars;
begin
  inherited PubliVars;
  PublicarVariableVR('Q', 'm3/s', 6, 2, Q, True, True);   //verificar unidades
end;

procedure TActorUniNodalCombustible.PrepararMemoria(globs: TGlobs);
begin
  inherited PrepararMemoria(globs);
  setlength(Q, globs.NPostes);
end;


procedure TActorUniNodalCombustible.Free;
begin
  setlength(Q, 0);
  inherited Free;
end;

//-------------------------
// Métodos de TActorBiNodalCombustible
//=========================

constructor TActorBiNodalCombustible.Create(capa: integer; nombre: string; nacimiento, muerte: TFecha;
  lpdUnidades: TFichasLPD; NodoCombA, NodoCombB: TNodoCombustible);
begin
  inherited Create(capa, nombre, nacimiento, muerte, lpdUnidades);
  self.NodoCombA := NodoCombA;
  self.NodoCombB := NodoCombB;
end;

constructor TActorBiNodalCombustible.Create_ReadFromText(f: TArchiTexto);
begin
  inherited Create_ReadFromText(f);
  f.IniciarLecturaRetrasada;
  f.rdReferencia('nodo combustible A', TCosa(NodoCombA), Self);
  f.rdReferencia('nodo combustible B', TCosa(NodoCombB), Self);
  f.EjecutarLectura;
end;

procedure TActorBiNodalCombustible.WriteToText(f: TArchiTexto);
begin
  inherited WriteToText(f);
  f.wrReferencia('nodo combustible A', NodoCombA);
  f.wrReferencia('nodo combustible B', NodoCombB);
end;

function TActorBiNodalCombustible.referenciaAlNodoCombustible (nodocomb: TNodoCombustible): boolean;
begin
  Result := (nodoCombA = nodocomb) or (nodoCombB = nodocomb);
end;

procedure TActorBiNodalCombustible.prepararMemoria(globs: TGlobs);
begin
  inherited prepararMemoria(globs);
  setlength(Q_NodoCombA, globs.NPostes);
  setlength(Q_NodoCombB, globs.NPostes);
end;

procedure TActorBiNodalCombustible.Free;
begin
  setlength(Q_NodoCombA, 0);
  setlength(Q_NodoCombB, 0);
  inherited Free;
end;

procedure AlInicio;
begin
  registrarClaseDeCosa(TActorUniNodalCombustible.ClassName, TActorUniNodalCombustible);
  registrarClaseDeCosa(TActorBiNodalCombustible.ClassName, TActorBiNodalCombustible);

end;

procedure AlFinal;
begin
end;

end.
