unit uTSumComConAlmacenYAgenda;

interface

uses
  uCosa, SysUtils, StdCtrls, uGlobs, uActores, uFechas, unodos, uCosaConNombre,
  uFichasLPD, usimplex, xMatDefs, uauxiliares, unodocombustible,
  uFuentesAleatorias, uTSumComb, uEstados, uAgendaGNL;

resourcestring
  rsSuministroCombustibleConAlmacenYAgenda =
    'Suministro Combustible Con Almacen y Agenda';


const
  factorRegas= 600; // (m3_GN / m3_GNL) este valore es 600


type
  TFichaSumComConAlmacenYAgenda = class(TFichaLPD)
    QMax: NReal;   // m3 de GN máximo SendOut de la regasificadora

    fd: NReal;
    TMR: NReal;

    VGNL_max: NReal; //volumen maximo del tanque
    VGNL_min: NReal; //volumen minimo de GNL que debe quedar en el tanque.

    ca_BOG: NReal; // BOG [m3/s] constante
    cb_BOG: NReal; // BOG [1/s] factor aplicable al volumen almacenado
    // QGNL_BOG:= ca_BOG + cb_BOG * X_GNL

    renRegas: NReal; // [p.u.] Rendimiento de regasificación.
    QGN_ConsumosPropios: NReal; // [m3/s] Consumos propios.

    constructor Create(
      capa: integer; fecha: TFecha; periodicidad: TPeriodicidad;
      QMax: NReal; fd, TMR: NReal;
      VGNL_max, VGNL_min: NReal;
      ca_BOG, cb_BOG, renRegas, QGN_ConsumosPropios: NReal);

    constructor Create_ReadFromText(f: TArchiTexto); override;
    procedure WriteToText(f: TArchiTexto); override;
    function InfoAd_: string; override;
  end;


  TSumComConAlmacenYAgenda = class(TSuministroSimpleCombustible)
  public

    // variables
    pa: TFichaSumComConAlmacenYAgenda;
    Agenda: TAgendaCargosGNL;


    //volumenes de GNL al inicio y al final del paso
    X_VGNL, Xs_VGNL: NReal; //Xs es final

    VGNL_inicial: NReal; //volumen inicial del tanque, antes de empezar la simulacion

    Ndiscret: integer;   //Nro de discretizaciones del tanque

    fuenteDeAporte_Agenda: TFuenteAleatoria; //descargas de GNL en el tanque
    nombreBorneAporte_Agenda: string;
    numeroBorneAporte_Agenda: integer;


    // faviables auxiliares
    cargo: TCargoGNL;
    V_Aportes: NReal; // Descargas de GNL en el paso de tiempo
    cv_GNL: NReal; // Valor futuro del GNL Almacenado
    Q_BOG, V_BOG: NReal; // Caudal de Boil Off Gas (BOG) y Volumen en el paso

    constructor Create(capa: integer; nombre: string; nacimiento, muerte: TFecha;
      lpdUnidades: TFichasLPD; lpd: TFichasLPD; NodoSumComb: TNodoCombustible;
      Agenda: TAgendaCargosGNL);

    constructor Create_ReadFromText(f: TArchiTexto); override;
    procedure WriteToText(f: TArchiTexto); override;
    procedure PubliVars; override;

    class function DescClase: string; override;

    procedure Sim_Cronica_Inicio; override;

    procedure RegistrarParametrosDinamicos; override;
    procedure optx_RegistrarVariablesDeEstado(adminEstados: TAdminEstados); override;
    procedure EvolucionarEstado; override;

    procedure PrepararMemoria(globs: TGlobs); override;
    procedure PrepararPaso_ps; override;

    procedure optx_nvxs(var ixr, ixd, iauxr, iauxd: integer); override;
    procedure ActualizarEstadoGlobal; override;
    procedure PosicionarseEnEstrellita; override;

    procedure opt_nvers(var ivar, ivae, ires: integer); override;
    procedure opt_cargue(s: TSimplex); override;
    procedure opt_fijarRestriccionesDeCaja(s: TSimplex); override;
    procedure opt_leerSolucion(s: TSimplex); override;



    procedure Free; override;




  end;

procedure AlInicio;
procedure ALFinal;

implementation



// Métodos de TSuministroCombustible

//function TSuministroCombustible.referenciaAlNodo(NodoSumComb:
//  TNodoCombustible): boolean;
//begin
//  Result := self.NodoCombustible =  NodoSumComb;
//end;

//constructor TSuministroCombustible.Create_ReadFromText(f: TArchiTexto);
//begin
//  inherited Create_ReadFromText( f );
//  f.rdReferencia( 'NodoCombustible', TCosa( NodoCombustible ), self );
//end;

//procedure TSuministroCombustible.WriteToText(f: TArchiTexto);
//begin
//  inherited WriteToText( f );
//  f.wrReferencia( 'NodoCombustible', Self.NodoCombustible );
//end;




// Métodos de TFichaSumComConAlmacenYAgenda

constructor TFichaSumComConAlmacenYAgenda.Create(
      capa: integer; fecha: TFecha; periodicidad: TPeriodicidad;
      QMax: NReal; fd, TMR: NReal;
      VGNL_max, VGNL_min: NReal;
      ca_BOG, cb_BOG, renRegas, QGN_ConsumosPropios: NReal);


begin
  inherited Create(capa, fecha, periodicidad);
  self.fd := fd;
  self.TMR := TMR;
  self.VGNL_max := VGNL_max;
  self.VGNL_min := VGNL_min;
  self.ca_BOG:= ca_BOG;
  self.cb_BOG:= cb_BOG;
  self.renRegas:= renRegas;
  self.QGN_ConsumosPropios:= QGN_ConsumosPropios;
end;

constructor TFichaSumComConAlmacenYAgenda.Create_ReadFromText(f: TArchiTexto);
begin
  inherited Create_ReadFromText(f);
  f.IniciarLecturaRetrasada;
  f.rd('QMax', QMax);
  f.rd('fd', fd);
  f.rd('TMR', TMR);
  f.rd('VGNL_max', VGNL_max);
  f.rd('VGNL_min', VGNL_min);
  f.rd('ca_BOG', ca_BOG);
  f.rd('cb_BOG', cb_BOG);
  f.rd('renRegas', renRegas);
  f.rd('QGN_ConsumosPropios', QGN_ConsumosPropios);
  f.EjecutarLectura;
end;

procedure TFichaSumComConAlmacenYAgenda.WriteToText(f: TArchiTexto);
begin
  inherited WriteToText(f);
  f.wr('QMax', QMax);
  f.wr('fd', fd);
  f.wr('TMR', TMR);
  f.wr('VGNL_max', VGNL_max);
  f.wr('VGNL_min', VGNL_min);
  f.wr('ca_BOG', ca_BOG);
  f.wr('cb_BOG', cb_BOG);
  f.wr('renRegas', renRegas);
  f.wr('QGN_ConsumosPropios', QGN_ConsumosPropios);
end;

function TFichaSumComConAlmacenYAgenda.InfoAd_: string;
begin
  Result := 'QMax: ' + FloatToStr(QMax) + ', VGNL_max:' + FloatToStr(VGNL_max);
end;



//--------------------------
// Métodos de TSumComConAlmacenYAgenda
//==========================

constructor TSumComConAlmacenYAgenda.Create(capa: integer; nombre: string;
  nacimiento, muerte: TFecha; lpdUnidades: TFichasLPD; lpd: TFichasLPD;
  NodoSumComb: TNodoCombustible;
  Agenda: TAgendaCargosGNL
  );

begin
  inherited Create(capa, nombre, nacimiento, muerte, lpdUnidades, lpd, NodoSumComb);
  self.Agenda := Agenda;
end;

constructor TSumComConAlmacenYAgenda.Create_ReadFromText(f: TArchiTexto);
begin
  inherited Create_ReadFromText(f);
  f.IniciarLecturaRetrasada;
  f.rd('Agenda', TCosa(agenda));
  f.EjecutarLectura;
end;

procedure TSumComConAlmacenYAgenda.WriteToText(f: TArchiTexto);
begin
  inherited WriteToText(f);
  f.wr('Agenda', agenda);
end;

procedure TSumComConAlmacenYAgenda.PubliVars;
begin
  inherited PubliVars;
end;



procedure TSumComConAlmacenYAgenda.PrepararMemoria(globs: TGlobs);
begin
  inherited PrepararMemoria(globs);
end;


class function TSumComConAlmacenYAgenda.DescClase: string;
begin
  Result := rsSuministroCombustibleConAlmacenYAgenda;
end;



procedure TSumComConAlmacenYAgenda.RegistrarParametrosDinamicos;
begin
  inherited registrarParametrosDinamicos;
  lpd.expandirFichas(globs);
  lpd.RegistrarFichasAActualizar(Self, globs.ActualizadorLPD, @pA, nil);
end;



procedure TSumComConAlmacenYAgenda.optx_RegistrarVariablesDeEstado(
  adminEstados: TAdminEstados);
begin
  adminEstados.Registrar_Continua(
    ixr,
    pa.VGNL_min,
    pa.VGNL_max,
    self.NDiscret,
    nombre + '_Vol', // nombre de la variable
    'm3' // unidades
    );
end;


procedure TSumComConAlmacenYAgenda.Sim_Cronica_Inicio;
begin
  X_VGNL := self.VGNL_inicial;
  Xs_VGNL := X_VGNL;
end;

procedure TSumComConAlmacenYAgenda.opt_nvers(var ivar, ivae, ires: integer);
begin
  self.ivar := ivar;
  Inc(ivar, globs.npostes + 1); //una ivar para cada Q por poste y una para el estado Xs.

  //la restricción es la ecuación de evolución del estado: Xs_VGNL = f(X,u,r)
  self.ires := ires;
  Inc(ires, 1);
end;


procedure TSumComConAlmacenYAgenda.PrepararPaso_ps;
var
  dCFdX_inc, dCFdX_dec: NReal;
  rescod: integer;
  xrpos: NReal;
  dCFdX: NReal;
begin
  inherited PrepararPaso_ps; // Calcula el PrecioOperativo aplicando la indexación y el precio
    // de la ficha LPD.


  // Ahora derivamos la función de CF(X, k+1) respecto al volumen almacenado
  if globs.EstadoDeLaSala = CES_OPTIMIZANDO then
    globs.CF.devxr_estrella(ixr, globs.kPaso_ + 1, dCFdX_inc, dCFdX_dec, rescod)
  else
    globs.CF.devxr_continuo(ixr, globs.kPaso_ + 1, dCFdX_inc,
      dCFdX_dec, rescod, xrpos);

  case rescod of
    -1: dCFdX := dCFdX_inc;
    0: dCFdX := (dCFdX_dec + dCFdX_inc) / 2.0;
    1: dCFdX := dCFdX_dec;
  end;

  // El vlor del GNL almacenado es menos la derivada del costo futuro respecto
  // al volumen almacenado y por el factor de actualización del paso.
  cv_GNL := -dCFdX * globs.fActPaso;

  // obtenemos una referencia al cargo activo
  Cargo:= Agenda.cargo( globs.kPaso_ );

  // El volumen de aportes es el Caudal Medio de aportes del paso
  // por la duración del paso de tiempo.
  V_Aportes:= Cargo.VGNL_m3;
  Q_BOG:= pa.ca_BOG + pa.cb_BOG * X_VGNL;
  V_BOG:= Q_BOG* globs.SegundosDelPaso; // Caudal de Boil Off Gas (BOG) y Volumen en el paso

end;

procedure TSumComConAlmacenYAgenda.opt_cargue(s: TSimplex);
var
  iresnodo: integer;
  kposte: integer;
  VAporte: NReal;
  dCFdV_Inc, dCFdV_Dec: NReal;
  resCode: integer;

begin
  iresnodo := self.NodoCombustible.ires;
  for kposte := 0 to globs.npostes - 1 do
  begin
    // En cada restricción del nodo_combustible aportamos
    // el caudal despachado.
    s.pon_e(iresnodo + kposte, ivar + kposte, 1);

    // Consideramos las extracciones asociadas a los caudales despachados
    // en la restricción dinámica f(X, u, r ) - Xs = 0
    // ( Aportes - SUM(Qi*durposi)  - Xs_VGNL + (X_VGNL - V_BOG) = 0)
    s.pon_e(ires, ivar + kposte, -globs.DurPos[kposte]);

    // cargar funcion objetivo -costo (variables Q por poste)
    s.pon_e(s.nf, ivar + kposte, -PrecioOperativo * globs.durpos[kposte]);
  end;

  //cargar termino de Xs_VGNL
  s.pon_e(ires, globs.npostes + 1, 1);

  //cargar termino constante
  s.pon_e(ires, s.nc, -self.X_VGNL + V_Aportes - V_BOG);


  // cargar funcion objetivo -costo (variable Aportes)
  s.acum_e(s.nf, s.nc, - cargo.precio * V_Aportes);


  // cargar funcion objetivo -costo (derivada de CF)
  s.pon_e(s.nf, ivar + globs.npostes + 1, dCFdV_Dec);
end;



procedure TSumComConAlmacenYAgenda.opt_fijarRestriccionesDeCaja(s: TSimplex);
var
  kposte: integer;
begin
  s.FijarRestriccionIgualdad(ires);

  for kposte := 0 to globs.npostes - 1 do
    s.cota_sup_set(ivar + kposte, QMaxOperativo);

  s.cota_sup_set(ivar + globs.npostes + 1, pa.VGNL_max);//  Xs_VGNL< VGNL_max
  s.cota_inf_set(ivar + globs.npostes + 1, pa.VGNL_min);//  Xs_VGNL> VGNL_min
end;



procedure TSumComConAlmacenYAgenda.EvolucionarEstado;
begin
  X_VGNL := Xs_VGNL;
end;

procedure TSumComConAlmacenYAgenda.optx_nvxs(var ixr, ixd, iauxr, iauxd: integer);
begin
  self.ixr := ixr;
  ixr := ixr + 1;
  Self.iauxr := iauxr;
  iauxr := iauxr + 1;
end;

procedure TSumComConAlmacenYAgenda.PosicionarseEnEstrellita;
begin
  X_VGNL := globs.CF.xr[ixr];
  Xs_VGNL := X_VGNL;
end;

procedure TSumComConAlmacenYAgenda.ActualizarEstadoGlobal;
begin
  globs.CF.xr[ixr] := X_VGNL;
end;


procedure TSumComConAlmacenYAgenda.opt_leerSolucion(s: TSimplex);
var
  kposte: integer;
  Consumo: NREal;
begin
  Consumo := 0;
  for kposte := 0 to globs.npostes - 1 do
  begin
    Q[kposte] := s.xval(ivar + kposte);
    Consumo := Consumo + Q[kposte] * globs.DurPos[kposte];
  end;
  self.X_VGNL := s.xval(ivar + globs.NPostes + 1);
  CostoDirectoDelPaso := Consumo * PrecioOperativo;
end;


procedure TSumComConAlmacenYAgenda.Free;
begin
  setlength(Q, 0);
  inherited Free;
end;

procedure AlInicio;
begin
  registrarClaseDeCosa(TSumComConAlmacenYAgenda.ClassName,
    TSumComConAlmacenYAgenda);
  registrarClaseDeCosa(TFichaSumComConAlmacenYAgenda.ClassName,
    TFichaSumComConAlmacenYAgenda);

end;

procedure AlFinal;
begin
end;

end.
