
(*
 rch.061122

 El propsito de esta unidad es definir una clase TAdminEstados
 que siva para administrar los estados de un sistema
 y brinde los servicios necesarios par realizar
 programacin dinmica estocstica (PDE) sobre  el sistema y
 realizar simulaciones del sistema una vez que se ha fijado
 la mejor poltica de operacin en base a la PDE.

 Para que el armado del espacio de estados se pueda realizar en forma
 colaborativa por el conjutno de Actores definimos mtodos para
 registrar variables de estado indicando el rango de la variable,
 el tipo (continua o discreta) y la cantidad de puntos de discretizacin
 que se desea tener sobre el rango. La mnima cantidad de puntos es 2 y
 seran los extremos del rango de la variable.

 Al inicio, en la creacin del sistema los Actores deben registar las variables
 Luego de registrados todos los actores, se debe indicar la cantidad de pasos
 de tiempo ms uno como la cantidad de FOTOS del espacio de estado que se
 deben considerar.

*)

unit uEstados;
{$MODE Delphi}

interface

uses
  Classes,
  Math, xmatdefs, uodt_types, uFechas, SysUtils, uCosa, uVersiones,
  ucontroladordeterminista;

resourcestring
  exCopiarVariablesDe =
    'TAdminEstados.copiarVariablesDe: el CF de destino no tiene la misma cantidad de variables que el CF de origen';


const
  cfbin_nver = 2; // agrego numero de versin al CF.bin

type
  TFuncTraduccion = function(x: NReal): NReal of object;

  TEngancheCFReal = class(TCosa)
  public
    nombreVar: string;
    unidades: string;
    valor: NReal;

    constructor Create(capa: integer; const nombreVar, unidades: string; valor: NReal);
    constructor Create_ReadFromText(f: TArchiTexto); override;
    procedure WriteToText(f: TArchiTexto); override;
  end;

  TDAOfEngancheCFReal = array of TEngancheCFReal;

  TEngancheCFEntero = class(TCosa)
  public
    nombreVar: string;
    unidades: string;
    valor: integer;

    constructor Create(capa: integer; const nombreVar, unidades: string; valor: integer);
    constructor Create_ReadFromText(f: TArchiTexto); override;
    procedure WriteToText(f: TArchiTexto); override;
  end;

  TDAOfEngancheCFEntero = array of TEngancheCFEntero;


  TAdminEstados = class
    // dimensionado
    nVarsContinuas: integer;
    nVarsDiscretas: integer;

    // vectores con las definiciones de las variablas continuas y discretas
    xr_def: TDAOfDefVarContinua;
    xd_def: TDAOfDefVarDiscreta;

    // descripcin de TIEMPO
    nPuntosT: integer; // nPasos+1
    fechaIni, fechaFin: TFecha;
    horasDelPaso: NReal;

    // soporte para el espacio de estado definido por los vectores anteriores
    constelacion: TConstelacion;

    // agrego soporte para controlador_determinista
    deterministico: boolean;
    // false si es un espacio de estado com, true si es para que sea determinista
    ControladorDeterministico: TControladorDeterministico;


    // Agregados para manejo de Trayectoria
    // vectores para almacenar el ESTADO ACTUAL
    xr: TDAOfNReal;
    xd: TDAOfNInt;

    // Indices de la posicin de la estrellita en cada  variables de estado
    // al recorrer las estrellas de un frame
    estrella_kr, estrella_kd: TDAOfNInt;

    // ordinal de la estrellita al recorrer las estrellas de un frame.
    ordinalEstrellaActual: integer;



    constructor Create(_nVarsContinuas, _nVarsDiscretas, _nPuntosT: integer);


    // carga un CF desde la posicin actual del archivo f  (formato binario)
    constructor Create_LoadFromFile(var f: file);
    // guarda CF desde la posicin actual del archivo f (formato binario)
    procedure StoreInFile(var f: file);

    // abre el archivo (archi) para lectura, llama a Create_LoadFromFile sobre
    // el archivo abierto y luego cierra (archi)
    constructor CreateLoadFromArchi(const archi: string);
    // abre el archivo (archi) para escritura, llama a StoreInFile y luego
    // cierra el archivo
    procedure StoreInArchi(const archi: string);

    procedure Free; virtual;

    procedure Registrar_Continua(ixr: integer; xmin, xmax: NReal;
      nPuntos: integer; nombre: string; unidades: string);

    function indexOf_Continua(nombre: string): integer;

    procedure RegistrarTraduccion_Continua(ixr: integer;
    // xmin, xmax: NReal;  nPuntos: integer;
      const nombre, unidades: string; funcTraduccion: TFuncTraduccion);

    procedure Registrar_Discreta(ixd: integer; npuntos: integer;
      const nombre, unidades: string);

    function indexOf_Discreta(nombre: string): integer;

    {
    procedure RegsitrarTraduccion_Discreta(ixd: integer;
    //npuntos: integer;
      nombre: string; unidades: string);
 }

    //Cuando cambiemos a paso variable cambiar el horasDelPaso por un array con
    //las horas de cada paso

    procedure CrearElEspacioTiempo(fechaIni, fechaFin: TFecha;
      horasDelPaso: NReal; const costoFuturo: TMatOfNReal; deterministico: boolean);


    procedure SetEstadoToEstrella;
    procedure SetEstadoToEstrella_indexada(vkr, vkd, idx_r, idx_d: TDAOfNInt);
    procedure SetEstrella_Indexada(vkr, vkd, idx_r, idx_d: TDAOfNInt);

    function costoEstrella(kpuntoT: integer): NReal;

    function costoContinuo(kPuntoT: integer): NReal;

    // interpona los vectores del Manto para la posicin continua actual y
    // carga el resultado en "res" desde la posicin jBase
    procedure mantoContinuo(var res: TDAOfNReal; jBase: integer;
      const Manto: TMatOfNReal);


    procedure SetCostoEstrella(kPuntoT: integer; valCosto: NReal);
    procedure AcumCostoEstrella(kPuntoT: integer; valCosto: NReal);

    // calcula la derivadas respecto de la variabla cuyo ndice es (irx) para
    // el punto de tiempo kpuntoT. La derivada dCdx_Inc es la calculada con
    // un incremento de x, la dCdx_Dec es calculada con un decremento.
    procedure devxr_estrella_20(irx: integer; kpuntoT: integer;
      var dCdx_Inc, dCdx_Dec: NReal);

    procedure devxr_continuo_20(irx: integer; kpuntoT: integer;
      var dCdx_Inc, dCdx_Dec: NReal);

    procedure devxr_estrella(irx: integer; kpuntoT: integer;
      var dCdx_Inc, dCdx_Dec: NReal; var resCod: integer);

    procedure devxr_continuo(irx: integer; kpuntoT: integer;
      var dCdx_Inc, dCdx_Dec: NReal; var resCod: integer;
    // -1 me fuir por abajo , 0 en la caja , 1 me fui por arriba
      var xrpos: NReal     // Posicin del punto en la cajita de paroximacin en por unidad
      );

    // Calcula la variacin del costo por un incremento delta_xd en la variable ird
    // en el punto de tiempo kpuntoT
    function deltaCosto_vxd_estrella(ird: integer; kpuntoT: integer;
      delta_xd: integer): NReal;
    function deltaCosto_vxd_continuo(ird: integer; kpuntoT: integer;
      delta_xd: integer): NReal;

    function deltaCosto_vxr_continuo_DosEstados_UTE(irx, irx2, kpuntoT: integer;
      delta_xr, delta_xr2: NReal): NReal;
    // delta costo sobre una coordenada real
    function deltaCosto_vxr_continuo(irx: integer; kpuntoT: integer;
      delta_xr: NReal): NReal;

    // fija los vectores ndices estrella_kr, estrella_kd al vrtice CERO
    // de la discretizacin.
    procedure setEstrellaCERO;

    // incrementa los ndices estrella_kr, estrella_kd de forma de apuntar
    // la siguiente estrella de la discretizacin.
    function IncEstrella: boolean;

    // Fija el punto de inicio de un barrido sobre los indexados;
    procedure SetEstrellaCERO_indexada(var estrella_idx_r, estrella_idx_d: TDAOfNint);

    // incrementa la estrella, pero sobre las variables indexadas
    function IncEstrella_indexada(var estrella_idx_r, estrella_idx_d: TDAOfNint;
      idx_r, idx_d: TDAOfNInt): boolean;

    procedure posicionarseEnEstrella(ordinalEstrella: integer);
    function nEstrellasPorPuntoT: integer;

    procedure InicializarFrameFinal(const archivoCF: string;
    // desde: TFecha;
      enganchesContinuos, enganchesDiscretos: TListaDeCosas;
      enganchar_promediando: boolean; uniformizar_promediando: string);

    // recibe un string con los nombres de las variables para uniformizar
    // promediando el ltimo frame.

    // Clasifica las variables en "a_Uniformizar" y "Otras".
    // entonces, hace un FOR en Otras recorriendo las estrellas que define
    // el producto cartesiano de sus discretizaciones. En cada Super_Estrella as definida
    // se recorren todas las estrellas definidas para ese punto del producto cartesinao de otras
    // por el producto cartesiano de las discretizaciones de las variables "a_Uniformizar".
    procedure uniformizar_ultimo_frame_en(uniformizar_promediando: string);


    function variablesIgualesA(otroCF: TAdminEstados): boolean;
    procedure copiarVariablesDe(otroCF: TAdminEstados);

    procedure dumpToTextFile(const nomArchivo: string);

    // hace un promedio del costo futuro de a nPasos.
    procedure Ecualizar(nPasos: integer);

    // Copiar el timo frame en todos los demas.
    procedure LlenarConFrameFinal;


    // retorna un string con las coordenadas de los estados continuos y discretos
    // separados por tabuladores.
    function GetEstado_XLT: string;

    // Cheque que el CFAux pasado como parmetro sea compatible para ser
    // accedido como un CFAux durante la simulacin.
    function ChequearCompatibilidad_CFAux( const CFAux: TAdminEstados ): boolean;
  end;

  (****OJO
  TAdminEstados_sobremuestreado = class( TAdminEstados )
    constructor
  end;
          ***)

procedure AlInicio;
procedure AlFinal;

implementation

constructor TEngancheCFReal.Create(capa: integer; const nombreVar, unidades: string;
  valor: NReal);
begin
  inherited Create(capa);
  self.nombreVar := nombreVar;
  self.unidades := unidades;
  self.valor := valor;
end;

constructor TEngancheCFReal.Create_ReadFromText(f: TArchiTexto);
var
  s: string;
  i1, i2: integer;
begin
  inherited Create_ReadFromText(f);
  f.IniciarLecturaRetrasada;
  f.rd('nombreVar', nombreVar);
  f.rd('valor', valor);
  if f.Version >= 94 then
    f.rd('unidades', unidades);
  f.EjecutarLectura;

  if f.Version < 94 then
  begin
    i1 := pos('[', nombreVar);
    i2 := pos(']', nombreVar);
    if (i1 > 0) and (i2 > i1) then
    begin
      unidades := trim(copy(nombreVar, i1 + 1, i2 - i1 - 1));
      Delete(nombreVar, i1, length(nombreVar) - i1 + 1);
      nombreVar := trim(nombreVar);
    end
    else
      unidades := '';
  end;
end;

procedure TEngancheCFReal.WriteToText(f: TArchiTexto);
begin
  inherited WriteToText(f);
  f.wr('nombreVar', nombreVar);
  f.wr('unidades', unidades);
  f.wr('valor', valor);
end;

constructor TEngancheCFEntero.Create(capa: integer; const nombreVar, unidades: string;
  valor: integer);
begin
  inherited Create(capa);
  self.nombreVar := nombreVar;
  self.unidades := unidades;
  self.valor := valor;
end;

constructor TEngancheCFEntero.Create_ReadFromText(f: TArchiTexto);
var
  i1, i2: integer;

begin
  inherited Create_ReadFromText(f);
  f.IniciarLecturaRetrasada;
  f.rd('nombreVar', nombreVar);
  if f.Version >= 94 then
    f.rd('unidades', unidades);

  f.rd('valor', valor);
  f.EjecutarLectura;

  if f.Version < 94 then
  begin
    i1 := pos('[', nombreVar);
    i2 := pos(']', nombreVar);
    if (i1 > 0) and (i2 > i1) then
    begin
      unidades := trim(copy(nombreVar, i1 + 1, i2 - i1 - 1));
      Delete(nombreVar, i1, length(nombreVar) - i1 + 1);
      nombreVar := trim(nombreVar);
    end
    else
      unidades := '';
  end;

end;

procedure TEngancheCFEntero.WriteToText(f: TArchiTexto);
begin
  inherited WriteToText(f);
  f.wr('nombreVar', nombreVar);
  f.wr('unidades', unidades);
  f.wr('valor', valor);
end;

procedure brwf(var f: file; var X: integer; flgw: boolean); overload;
begin
  if flgw then
    BlockWrite(f, X, SizeOf(X))
  else
    BlockRead(f, X, SizeOf(X));
end;

procedure brwf(var f: file; var X: NReal; flgw: boolean); overload;
begin
  if flgw then
    BlockWrite(f, X, SizeOf(X))
  else
    BlockRead(f, X, SizeOf(X));
end;

procedure brwf(var f: file; var X: TDateTime; flgw: boolean); overload;
begin
  if flgw then
    BlockWrite(f, X, SizeOf(X))
  else
    BlockRead(f, X, SizeOf(X));
end;

procedure brwf(var f: file; var X: TDAOfNReal; flgw: boolean); overload;
var
  n, k: integer;
begin
  if flgw then
  begin
    n := length(X);
    brwf(f, n, True);
    for k := 0 to high(X) do
      brwf(f, x[k], True);
  end
  else
  begin
    brwf(f, n, False);
    setlength(X, n);
    for k := 0 to high(X) do
      brwf(f, x[k], False);
  end;
end;

procedure brwf(var f: file; var X: TDAOfNInt; flgw: boolean); overload;
var
  n, k: integer;
begin
  if flgw then
  begin
    n := length(X);
    brwf(f, n, True);
    for k := 0 to high(X) do
      brwf(f, x[k], True);
  end
  else
  begin
    brwf(f, n, False);
    setlength(X, n);
    for k := 0 to high(X) do
      brwf(f, x[k], False);
  end;
end;

procedure brwf(var f: file; var x: TDAOfDefVarContinua; flgw: boolean);
  overload;
var
  n, k: integer;
begin
  if flgw then
  begin
    n := length(X);
    brwf(f, n, True);
    for k := 0 to high(X) do
      x[k].StoreInFile(f);
  end
  else
  begin
    brwf(f, n, False);
    setlength(X, n);
    for k := 0 to high(X) do
      x[k] := TDefVarContinua.Create_LoadFromFile(f);
  end;
end;

procedure brwf(var f: file; var x: TDAOfDefVarDiscreta; flgw: boolean);
  overload;
var
  n, k: integer;
begin
  if flgw then
  begin
    n := length(X);
    brwf(f, n, True);
    for k := 0 to high(X) do
      x[k].StoreInFile(f);
  end
  else
  begin
    brwf(f, n, False);
    setlength(X, n);
    for k := 0 to high(X) do
      x[k] := TDefVarDiscreta.Create_LoadFromFile(f);
  end;
end;

procedure TAdminEstados.StoreInArchi(const archi: string);
var
  f: file;
begin
  assignFile(f, archi);
  rewrite(f, blockSize_ArchiBin);
  StoreInFile(f);
  closefile(f);
end;

procedure TAdminEstados.StoreInFile(var f: file);
var
  flgw: boolean;
  raux: double;
  iaux: integer;

begin
  flgw := True;
  raux:= -1;
  iaux:= cfbin_nver;
  brwf(f, raux, flgw);
  brwf(f, iaux, flgw);
  brwf(f, fechaIni.dt, flgw);
  brwf(f, fechaFin.dt, flgw);
  brwf(f, horasDelPaso, flgw);
  brwf(f, nVarsContinuas, flgw);
  brwf(f, nVarsContinuas, flgw);
  brwf(f, nVarsDiscretas, flgw);
  brwf(f, nPuntosT, flgw);
  brwf(f, xr_def, flgw);
  brwf(f, xd_def, flgw);
  constelacion.StoreInFile(f);
  brwf(f, xr, flgw);
  brwf(f, xd, flgw);
  brwf(f, estrella_kr, flgw);
  brwf(f, estrella_kd, flgw);
  brwf(f, ordinalEstrellaActual, flgw);
end;

function TAdminEstados.variablesIgualesA(otroCF: TAdminEstados): boolean;
var
  res: boolean;
  i: integer;
begin
  if (nVarsContinuas = otroCF.nVarsContinuas) or (nVarsDiscretas =
    otroCF.nVarsDiscretas) then
  begin
    res := True;
    for i := 0 to nVarsContinuas - 1 do
      if not xr_def[i].esIgualA(otroCF.xr_def[i]) then
      begin
        res := False;
        break;
      end;

    if res then
      for i := 0 to nVarsContinuas - 1 do
        if not xd_def[i].esIgualA(otroCF.xd_def[i]) then
        begin
          res := False;
          break;
        end;
    Result := res;
  end
  else
    Result := False;
end;

procedure TAdminEstados.copiarVariablesDe(otroCF: TAdminEstados);
var
  i: integer;
begin
  if (Self.nVarsContinuas <> otroCF.nVarsContinuas) or
    (self.nVarsDiscretas <> otroCF.nVarsDiscretas) then
    raise Exception.Create(exCopiarVariablesDe);

  for i := 0 to High(otroCF.xr_def) do
    self.xr_def[i] := otroCF.xr_def[i].clonar;
  for i := 0 to high(otroCF.xd_def) do
    self.xd_def[i] := otroCF.xd_def[i].clonar;
end;

constructor TAdminEstados.CreateLoadFromArchi(const archi: string);
var
  f: file;
begin
  assignFile(f, archi);
{$I-}
  reset(f, blockSize_ArchiBin);
{$I+}
  if ioresult <> 0 then
    raise Exception.Create('Error, no pude abrir el archivo: ' + archi);

  Create_LoadFromFile(f);
  closefile(f);
end;

constructor TAdminEstados.Create_LoadFromFile(var f: file);
var
  flgw: boolean;
  dtAux: TDateTime;
  v0: double; // primer numero. Si es -1 indica que
    // el segundo es el nmero de versin.
    // si no es -1 asumo que el numero de versin es 1
  nver: integer;
  intaux: integer;

begin
  flgw := False;
  brwf(f, v0, flgw);
  if ( v0 = -1 ) then
  begin
    brwf( f, nver, flgw );
    brwf(f, dtAux, flgw );
  end
  else
  begin
    nver:= 1;
    dtAux:= TDateTime( v0 );
  end;

  FechaIni := TFecha.Create_Dt(dtAux);
  brwf(f, dtAux, flgw);
  fechaFin := TFecha.Create_Dt(dtAux);

  if nver < 2 then
  begin
    brwf(f, intaux, flgw);
    horasDelPaso:= intaux;
  end
  else
  begin
    brwf(f, horasDelPaso, flgw);
  end;
  brwf(f, nVarsContinuas, flgw);
  brwf(f, nVarsContinuas, flgw);
  brwf(f, nVarsDiscretas, flgw);
  brwf(f, nPuntosT, flgw);
  brwf(f, xr_def, flgw);
  brwf(f, xd_def, flgw);
  constelacion := TConstelacion.Create_LoadFromFile(f);
  brwf(f, xr, flgw);
  brwf(f, xd, flgw);
  brwf(f, estrella_kr, flgw);
  brwf(f, estrella_kd, flgw);
  brwf(f, ordinalEstrellaActual, flgw);
end;

function TAdminEstados.IncEstrella: boolean;
begin
  Result := constelacion.inc_estrella(estrella_kr, estrella_kd);
  ordinalEstrellaActual := constelacion.ordinalEstrella(estrella_kr, estrella_kd);
end;


function TAdminEstados.IncEstrella_indexada(
  var estrella_idx_r, estrella_idx_d: TDAOfNint; idx_r, idx_d: TDAOfNInt): boolean;

begin
  Result := constelacion.inc_estrella_indexada(estrella_idx_r,
    estrella_idx_d, idx_r, idx_d);
end;




procedure TAdminEstados.posicionarseEnEstrella(ordinalEstrella: integer);
begin
  constelacion.indicesEstrella_i(estrella_kr, estrella_kd, ordinalEstrella);
  ordinalEstrellaActual := ordinalEstrella;
end;

function TAdminEstados.nEstrellasPorPuntoT: integer;
begin
  Result := constelacion.nEstrellas;
end;



procedure TAdminEstados.InicializarFrameFinal(const archivoCF: string;
  //  desde: TFecha;
  enganchesContinuos, enganchesDiscretos: TListaDeCosas;
  enganchar_promediando: boolean; uniformizar_promediando: string);

var
  //El costo futuro con los valores que voy a copiar
  CFaux: TAdminEstados;
  //Guarda los indices en self de las variables comunes a ambos CFs
  xr_varsComunes, xd_varsComunes: TDAofNInt;
  //Guarda los indices en CFaux de las variables comunes a ambos CFs
  xr_varsComunesAux, xd_varsComunesAux: TDAofNInt;
  costoAux: NReal;

  desde: TFecha;

  i, j, nVarsComunes: integer;
  finBarridoEstrellas: boolean;
  kpr, alfa, beta: NReal;
  kp1, kp2: integer;
  encontre: boolean;
{$IFDEF DBG}
  fDbg: Textfile;
{$ENDIF}
// Desaparecida significa que Est En CFAux y no en Self.
  cnt_desaparecidas_continuas, cnt_desaparecidas_discretas: integer;
  k_desaparecidas_continuas, k_desaparecidas_discretas: TDAOfNInt;
  cnt_: integer;
  encontradas: array of boolean;
  contadores_continuas, contadores_discretas: TDAOfNInt;
  k: integer;
  recorriendo_desaparecidas: boolean;
  hay_desaparecidas: boolean;

  v_uniformizar: TStrings;

begin
  desde := self.fechaFin;

  try
    CFaux := TAdminEstados.CreateLoadFromArchi(archivoCF);
  except
    raise Exception.Create('Error al intentar leer CFaux: ' + archivoCF +
      '. Capaz deba optimizar la otra sala primero.');
  end;

  //kpr es la cantidad de pasos desde el inicio de CFaux hasta "desde.dt"
  //con el paso de tiempo de aux
  kpr := (desde.dt - CFaux.fechaIni.dt) * 24.0 / CFaux.horasDelPaso + 1;
  alfa := frac(kpr);
  beta := 1 - alfa;
  kp1 := trunc(kpr);
  kp2 := kp1 + 1;

  if (kp1 <= 0) or (kp2 > high(CFAux.constelacion.fcosto)) then
    raise Exception.Create(
      'TAdminEstados.InicializarFrameFinal: se desea enganchar el ' +
      desde.AsStr + ' y el archivo de costo futuro con que desea enganchar va desde el '
      + CFaux.fechaIni.AsStr + ' hasta el ' + CFaux.fechaFin.AsStr +
      '.'#10'La fecha de enganche debe estar contenida dentro del perodo de optimizacin del archivo CF.bin');

  //Identifico las variables reales comunes a ambos CFs
  setLength(xr_varsComunes, length(xr_def));
  setLength(xr_varsComunesAux, length(CFaux.xr_def));
  setlength(encontradas, length(CFaux.xr_def));
  nVarsComunes := 0;
  cnt_desaparecidas_continuas := length(CFaux.xr_def);
  for j := 0 to high(CFaux.xr_def) do
  begin  // buscamos en CFaux si hay una con igual nombre
    encontradas[j] := false;
    for i := 0 to high(xr_def) do
    begin // para cada variable del CF
      if xr_def[i].nombre = CFaux.xr_def[j].nombre then
      begin
        xr_varsComunes[nVarsComunes] := i;
        xr_varsComunesAux[nVarsComunes] := j;
        Inc(nVarsComunes);
        Dec(cnt_desaparecidas_continuas);
        encontradas[j] := True;
        break;
      end;
    end;
  end;

  setlength(k_desaparecidas_continuas, cnt_desaparecidas_continuas);
  cnt_ := 0;
  for j := 0 to high(CFaux.xr_def) do
    if not encontradas[j] then
    begin
      k_desaparecidas_continuas[cnt_] := j;
      Inc(cnt_);
    end;


  if nVarsComunes < Length(xr_varsComunes) then
    if nVarsComunes <> 0 then
      xr_varsComunes := copy(xr_varsComunes, 0, nVarsComunes)
    else
      xr_varsComunes := nil;

  if nVarsComunes < Length(xr_varsComunesAux) then
    if nVarsComunes <> 0 then
      xr_varsComunesAux := copy(xr_varsComunesAux, 0, nVarsComunes)
    else
      xr_varsComunesAux := nil;

  cnt_desaparecidas_discretas := length(CFaux.xd_def);
  setlength(encontradas, length(CFaux.xd_def));
  //Identifico las variables discretas comunes a ambos CFs
  SetLength(xd_varsComunes, length(xd_def));
  SetLength(xd_varsComunesAux, length(CFaux.xd_def));
  nVarsComunes := 0;
  for j := 0 to high(CFaux.xd_def) do
  begin
    encontradas[j] := False;
    for i := 0 to high(xd_def) do
    begin
      if xd_def[i].nombre = CFaux.xd_def[j].nombre then
      begin
        xd_varsComunes[nVarsComunes] := i;
        xd_varsComunesAux[nVarsComunes] := j;
        Inc(nVarsComunes);
        Dec(cnt_desaparecidas_discretas);
        encontradas[j] := True;
        break;
      end;
    end;
  end;

  setlength(k_desaparecidas_discretas, cnt_desaparecidas_discretas);
  cnt_ := 0;
  for j := 0 to high(CFaux.xd_def) do
    if not encontradas[j] then
    begin
      k_desaparecidas_discretas[cnt_] := j;
      Inc(cnt_);
    end;


  if nVarsComunes < Length(xd_varsComunes) then
    if nVarsComunes <> 0 then
      xd_varsComunes := copy(xd_varsComunes, 0, nVarsComunes)
    else
      xd_varsComunes := nil;

  if nVarsComunes < Length(xd_varsComunesAux) then
    if nVarsComunes <> 0 then
      xd_varsComunesAux := copy(xd_varsComunesAux, 0, nVarsComunes)
    else
      xd_varsComunesAux := nil;

  hay_desaparecidas := (cnt_desaparecidas_continuas + cnt_desaparecidas_discretas) > 0;

  setEstrellaCERO; // iniciamos barrido

  //Resolver los enganches
  //Fijamos los valores especificados de las variables continuas en el CFAux
  for j := 0 to high(CFaux.xr_def) do
  begin
    encontre := False;
    for i := 0 to enganchesContinuos.Count - 1 do
    begin
      if CFaux.xr_def[j].nombre = TEngancheCFReal(enganchesContinuos[i])
        .nombreVar then
      begin
        CFaux.xr[j] := TEngancheCFReal(enganchesContinuos[i]).valor;
        encontre := True;
        break;
      end;
    end;
    if not encontre then  //Si no me especificaron su valor, la pongo en el valor medio
      CFaux.xr[j] := CFAux.xr_def[j].xmed;
  end;

  //Fijamos los valores especificados de las variables discretas en el CFAux
  for j := 0 to high(CFaux.xd_def) do
  begin
    encontre := False;
    for i := 0 to enganchesDiscretos.Count - 1 do
    begin
      if CFaux.xd_def[j].nombre = TEngancheCFEntero(enganchesDiscretos[i])
        .nombreVar then
      begin
        CFaux.xd[j] := TEngancheCFEntero(enganchesDiscretos[i]).valor;
        encontre := True;
        break;
      end;
    end;
    if not encontre then  //Si no me especificaron su valor, la pongo en el valor mnimo
      CFaux.xd[j] := CFAux.xd_def[j].x[0];
  end;

{$IFDEF DBG}
  AssignFile(fDbg, 'inicializarFrameFinalDbg.xlt');
  rewrite(fDbg);
{$ENDIF}

  if hay_desaparecidas and enganchar_promediando then
  begin
    // si se va a precisar iterar en los estados perdidos de CFAux ponemos
    // los contadores de las discretizaciones para poder iterar.
    setlength(contadores_continuas, cnt_desaparecidas_continuas);
    setlength(contadores_discretas, cnt_desaparecidas_discretas);
    for k := 0 to high(contadores_continuas) do
      contadores_continuas[k] := CFaux.xr_def[k_desaparecidas_continuas[k]].NPuntos;
    for k := 0 to high(contadores_discretas) do
      contadores_continuas[k] := CFaux.xd_def[k_desaparecidas_discretas[k]].NPuntos;
  end;

  repeat
    //Iteramos en los estados de CF
    SetEstadoToEstrella; // cargamos las variables de acuerdo con la estrella

    for i := 0 to High(xr_varsComunes) do
      CFaux.xr[xr_varsComunesAux[i]] := xr[xr_varsComunes[i]];
    for i := 0 to High(xd_varsComunes) do
      CFaux.xd[xd_varsComunesAux[i]] := xd[xd_varsComunes[i]];

    if not (hay_desaparecidas and enganchar_promediando) then
    begin
      if abs(alfa) > AsumaCero then
        costoAux := CFaux.costoContinuo(kp1) * beta +
          CFaux.costoContinuo(kp2) * alfa
      else
        costoAux := CFaux.costoContinuo(kp1);
    end
    else
    begin
      costoAux := 0;
      cnt_ := 0;

      recorriendo_desaparecidas := True;
      CFaux.SetEstrellaCERO_indexada(contadores_continuas, contadores_discretas);

      while recorriendo_desaparecidas do
      begin
        // fijamos los valores de las indexaddas
        for i := 0 to high(k_desaparecidas_continuas) do
          CFaux.xr[k_desaparecidas_continuas[i]] := contadores_continuas[i];
        for i := 0 to high(k_desaparecidas_discretas) do
          CFaux.xr[k_desaparecidas_discretas[i]] := contadores_discretas[i];

        if abs(alfa) > AsumaCero then
          costoAux := costoAux + CFaux.costoContinuo(kp1) * beta +
            CFaux.costoContinuo(kp2) * alfa
        else
          costoAux := costoAux + CFaux.costoContinuo(kp1);

        Inc(cnt_);

        recorriendo_desaparecidas :=
          CFAux.IncEstrella_indexada(contadores_continuas,
          contadores_discretas, k_desaparecidas_continuas, k_desaparecidas_discretas);

      end;
      costoAux := costoAux / cnt_;
    end;
    constelacion.set_costo_estrella(high(constelacion.fcosto),
      ordinalEstrellaActual, costoAux);
{$IFDEF DBG}
    Write(fDbg, FloatToStrF(costoAux, ffGeneral, 6, 2), #9);
{$ENDIF}
    finBarridoEstrellas := not IncEstrella;
  until (finBarridoEstrellas);
{$IFDEF DBG}
  CloseFile(fDbg);
{$ENDIF}

  // Ahora si hay alguna en uniformizar_promediando que
  // no haya desaparecido calculo el promedio y se lo impongo
  uniformizar_ultimo_frame_en(uniformizar_promediando);
end;


procedure TAdminEstados.uniformizar_ultimo_frame_en(uniformizar_promediando: string);
var
  v_uniformizar: TStringList;
  clasificadas_r, clasificadas_d: array of boolean;
  h, i, j, k: integer;
  cnt_r: integer;
  cnt_d: integer;

  a_uniformizar_r, otras_r, a_uniformizar_d, otras_d: TDAofNInt;
  u_r, u_d, o_r, o_d: TDAOfNInt;
  loop_otras: boolean;
  m: NREal;
  loop_uniformizar: boolean;
  cnt_: integer;
begin
  if trim(uniformizar_promediando) = '' then
    exit;

  // creamos una lista con el nombre de las variables
  v_uniformizar := TStringList.Create;
  v_uniformizar.Delimiter := ';';
  v_uniformizar.DelimitedText := uniformizar_promediando;

  if v_uniformizar.Count = 0 then
    exit;

  // armo el grupo de los a_uniformizar y el grupo otras
  setlength(clasificadas_r, length(xr_def));
  setlength(clasificadas_d, length(xd_def));

  for k := 0 to high(clasificadas_r) do
    clasificadas_r[k] := False;
  for k := 0 to high(clasificadas_d) do
    clasificadas_d[k] := False;

  cnt_r := 0;
  cnt_d := 0;

  for k := 0 to v_uniformizar.Count - 1 do
  begin
    j := indexOf_Continua(v_uniformizar[k]);
    if j >= 0 then
    begin
      clasificadas_r[j] := True;
      Inc(cnt_r);
    end
    else
    begin
      j := indexOf_Discreta(v_uniformizar[k]);
      if j < 0 then
        raise Exception.Create('No enonctr: ' + v_uniformizar[k] +
          ', en la funcin de CF. Imposible aplicar uniformizador sobre dicha variable');
      clasificadas_d[j] := True;
      Inc(cnt_d);
    end;
  end;

  setlength(a_uniformizar_r, cnt_r);
  setlength(otras_r, length(xr_def) - cnt_r);

  j := 0;
  h := 0;
  for k := 0 to high(clasificadas_r) do
    if clasificadas_r[k] then
    begin
      a_uniformizar_r[j] := k;
      Inc(j);
    end
    else
    begin
      otras_r[h] := k;
      Inc(h);
    end;

  setlength(a_uniformizar_d, cnt_d);
  setlength(otras_d, length(xd_def) - cnt_d);
  j := 0;
  h := 0;
  for k := 0 to high(clasificadas_d) do
    if clasificadas_d[k] then
    begin
      a_uniformizar_d[j] := k;
      Inc(j);
    end
    else
    begin
      otras_d[h] := k;
      Inc(h);
    end;

  setlength(o_r, length(otras_r));
  setlength(o_d, length(otras_d));
  setlength(u_r, length(a_uniformizar_r));
  setlength(u_d, length(a_uniformizar_d));

  SetEstrellaCERO_indexada(o_r, o_d);
  loop_otras := True;
  while loop_otras do
  begin
    SetEstrella_Indexada(o_r, o_d, otras_r, otras_d);

    // calculamos el promedio
    m := 0;
    SetEstrellaCERO_indexada(u_r, u_d);
    loop_uniformizar := True;
    cnt_ := 0;
    while loop_uniformizar do
    begin
      SetEstrella_Indexada(u_r, u_d, a_uniformizar_r, a_uniformizar_d);
      m := m + costoEstrella(high(constelacion.fcosto));
      Inc(cnt_);
      loop_uniformizar := IncEstrella_indexada(u_r, u_d, a_uniformizar_r,
        a_uniformizar_d);
    end;
    m := m / cnt_;

    // imponemos el promedio calculado
    SetEstrellaCERO_indexada(u_r, u_d);
    loop_uniformizar := True;
    cnt_ := 0;
    while loop_uniformizar do
    begin
      SetEstrella_Indexada(u_r, u_d, a_uniformizar_r, a_uniformizar_d);
      SetCostoEstrella(high(constelacion.fCosto), m);
      loop_uniformizar := IncEstrella_indexada(u_r, u_d, a_uniformizar_r,
        a_uniformizar_d);
    end;


    loop_otras := IncEstrella_Indexada(o_r, o_d, otras_r, otras_d);
  end;

end;

procedure TAdminEstados.setEstrellaCERO;
begin
  vclear(estrella_kr);
  vclear(estrella_kd);
  ordinalEstrellaActual := 0;
end;

procedure TAdminEstados.SetEstrellaCERO_indexada(
  var estrella_idx_r, estrella_idx_d: TDAOfNint);
begin
  vclear(estrella_idx_r);
  vclear(estrella_idx_d);
end;


procedure TAdminEstados.SetEstrella_Indexada(vkr, vkd, idx_r, idx_d: TDAOfNInt);
var
  i: integer;
begin
  for i := 0 to high(idx_r) do
    estrella_kr[idx_r[i]] := vkr[i];
  for i := 0 to high(idx_d) do
    estrella_kd[idx_d[i]] := vkd[i];
end;


procedure TAdminEstados.AcumCostoEstrella(kPuntoT: integer; valCosto: NReal);
begin
  constelacion.acum_costo_estrella(kPuntoT, estrella_kr, estrella_kd, valCosto);
  //  constelacion.acum_costo_estrella( kPuntoT, ordinalEstrellaActual, valCosto );
end;

procedure TAdminEstados.SetCostoEstrella(kPuntoT: integer; valCosto: NReal);
begin
  constelacion.set_costo_estrella(kPuntoT, estrella_kr, estrella_kd, valCosto);
  //  constelacion.set_costo_estrella( kPuntoT, ordinalEstrellaActual, valCosto );
end;

function TAdminEstados.costoEstrella(kpuntoT: integer): NReal;
begin
  Result := constelacion.costo_estrella(kpuntoT, estrella_kr, estrella_kd);
  //  result:= constelacion.costo_estrella(kpuntoT, ordinalEstrellaActual);
end;

function TAdminEstados.costoContinuo(kPuntoT: integer): NReal;
begin
  if deterministico then
    if kPuntoT < nPuntosT then
      Result := ControladorDeterministico.GradInfos[kPuntoT - 1].CF
    else
      Result := constelacion.costo_continuo(1, xr, xd)
  else
    Result := constelacion.costo_continuo(kpuntoT, xr, xd);
end;

function TAdminEstados.GetEstado_XLT: string;
var
  k: integer;
  res: string;
begin
  res := 'xr:';
  for k := 0 to high(xr) do
    res := res + #9 + FloatToStr(xr[k]);
  res := res + #9 + 'xd:';
  for k := 0 to high(xd) do
    res := res + #9 + IntToStr(xd[k]);
  Result := res;
end;

function TAdminEstados.ChequearCompatibilidad_CFAux( const CFAux: TAdminEstados ): boolean;
var
  msg: string;


  procedure apm( s:string );
  begin
    if msg <> '' then
     msg:= msg + ', ';
    msg:= msg + s;
  end;

begin
  msg:= '';
  // chequear ventana temporal
  if nPuntosT <> CFAux.nPuntosT then
    apm( 'nPuntosT <> CFAux.nPuntosT' );

  if fechaIni.dt <> CFAux.fechaIni.dt then
    apm( 'fechaIni.dt <> CFAux.fechaIni.dt' );

  if fechaFin.dt <> CFAux.fechaFin.dt then
    apm( 'fechaFin.dt <> CFAux.fechaFin.dt' );

  if horasDelPaso <> CFAux.horasDelPaso then
    apm( 'horasDelPaso <> CFAux.horasDelPaso' );


  // dimensionado
   if  nVarsContinuas <> CFAux.nVarsContinuas then
     apm( 'nVarsContinuas <> CFAux.nVarsContinuas' ) ;
   if  nVarsDiscretas <> CFAux.nVarsDiscretas then
     apm( 'nVarsDiscretas <> CFAux.nVarsDiscretas' ) ;

   if msg <> '' then
   begin
     result:= false;
     raise Exception.Create('CFAux incompatible!!! : '+msg );

   end
   else
    result:= true;
end;

procedure TAdminEstados.mantoContinuo(var res: TDAOfNReal; jBase: integer;
  const Manto: TMatOfNReal);
begin
  constelacion.manto_continuo(xr, xd, res, jBase, Manto);
end;

procedure TAdminEstados.devxr_estrella(irx: integer; kpuntoT: integer;
  var dCdx_Inc, dCdx_Dec: NReal; var resCod: integer);
begin
  constelacion.dev_costo_estrella(kpuntoT, estrella_kr, estrella_kd, irx,
    dCdx_Inc, dCdx_Dec, resCod);
end;

procedure TAdminEstados.devxr_estrella_20(irx: integer; kpuntoT: integer;
  var dCdx_Inc, dCdx_Dec: NReal);
var
  codRes: integer;
  //  xrpos: NReal;
begin
  devxr_estrella(irx, kpuntoT, dCdx_Inc, dCdx_Dec, codRes);

  if codRes <> 0 then
  begin
    if codRes < 0 then
      dCdx_Dec := dCdx_Inc * 2
    else
      dCdx_Inc := 0;
  end;
end;

procedure TAdminEstados.dumpToTextFile(const nomArchivo: string);
var
  f: Textfile;
  k, j: integer;
  iterFecha, dtDelPaso: TDateTime;
begin
  assignFile(f, nomArchivo);
  try
    rewrite(f);
    dtDelPaso := horasDelPaso * horaToDt;

    Writeln(f, 'Versin del simulador:'#9, vSimSEESimulador);
    Writeln(f, 'fActPaso: '#9, '???'
      {FloatToStrF(globs.fActPaso, ffGeneral, 12, 10 )});
    constelacion.PrintDefsToText(f, True);
    Write(f, 'paso\estado'#9'Fecha');
    for k := 1 to nEstrellasPorPuntoT do
      Write(f, #9, k);
    Writeln(f);

    // salvamos el frame del ltimo paso de tiempo
    for k := nPuntosT - 1 downto 1 do
    begin
      iterFecha := fechaIni.dt + dtDelPaso * (k - 1);
      Write(f, IntToStr(k) + #9 + DateTimeToIsoStr(iterFecha));

      for j := 0 to nEstrellasPorPuntoT - 1 do
        Write(f, #9, FloatToStrF(constelacion.fcosto[k][j], ffGeneral, 6, 2));
      Writeln(f);
    end;

  finally
    closefile(f);
  end;
end;

// hace un promedio del costo futuro de a nPasos.
procedure TAdminEstados.Ecualizar(nPasos: integer);
var
  k, k_ultimo, dk, j: integer;
  a: NReal;
begin
  if nPasos <= 0 then
    raise Exception.Create('TAdminEstados.Ecualizar nPasos <= 0 ');

  k_ultimo := nPuntosT - nPasos;
  if k_ultimo <= 0 then
  begin
    k_ultimo := 1;
    nPasos := nPuntosT;
  end;

  for k := 1 to k_ultimo do
  begin
    for j := 0 to nEstrellasPorPuntoT - 1 do
    begin
      a := 0;
      for dk := 0 to nPasos - 1 do
        a := a + constelacion.fcosto[k + dk][j];
      a := a / nPasos;
      constelacion.fcosto[k][j] := a;
    end;
  end;
  for k := k_ultimo + 1 to nPuntosT - 1 do
    for j := 0 to nEstrellasPorPuntoT - 1 do
      constelacion.fcosto[k][j] := constelacion.fcosto[k - 1][j];
end;

procedure TAdminEstados.LlenarConFrameFinal;
var
  k, j: integer;
  a: NReal;
begin
  // k puede ir de 1 a nPuntosT-1
  // k= 1nPuntosT-1 corresponde al fin de la ltima etapa.
  // k= 1 corresponde al inicio de la primer etapa
  // k= 0 no se utiliza

  // para cada estrella
  for j := 0 to nEstrellasPorPuntoT - 1 do
  begin
    // valor en el ltimo frame
    a := constelacion.fcosto[nPuntosT - 1][j];
    for k := 1 to nPuntosT - 2 do
      constelacion.fcosto[k][j] := a;
    // lo copio en las estrellas de los frames anteriores
  end;
end;

procedure TAdminEstados.devxr_continuo(irx: integer; kpuntoT: integer;
  var dCdx_Inc, dCdx_Dec: NReal; var resCod: integer;
  // -1 me fuir por abajo , 0 en la caja , 1 me fui por arriba
  var xrpos: NReal     // Posicin del punto en la cajita de paroximacin en por unidad
  );
begin
  if deterministico then
  begin
    if kPuntoT < nPuntosT then
    begin
      dCdx_Inc := ControladorDeterministico.GradInfos[kPuntoT - 1].Grad_Xr[irx];
      dCdx_Dec := dcdx_Inc;
      resCod := 0;
      xrpos := 0;
    end
    else
      constelacion.dev_costo_continuo(1, xr, xd, irx, dCdx_Inc, dCdx_Dec,
        resCod, xrpos);
  end
  else
    constelacion.dev_costo_continuo(kpuntoT, xr, xd, irx, dCdx_Inc, dCdx_Dec,
      resCod, xrpos);
end;


procedure TAdminEstados.devxr_continuo_20(irx: integer; kpuntoT: integer;
  var dCdx_Inc, dCdx_Dec: NReal);

var
  codRes: integer;
  xrpos: NReal;

begin
  devxr_continuo(irx, kpuntoT, dCdx_Inc, dCdx_Dec, codRes, xrpos);

  if codRes <> 0 then
  begin
    if codRes < 0 then
    begin
      if xrpos > 0 then
      begin
        dCdx_Dec := dCdx_Dec * xrpos + (1 - xrpos) * dCdx_Inc * 2;

        // me aseguro la convexidad.
        dCdx_Dec := min(dCdx_Inc, dCdx_Dec);
      end
      else
        dCdx_Dec := dCdx_Inc * 2;
    end
    else
    begin
      if xrpos < 1 then
      begin
        // me aseguro la convexidad
        dCdx_Inc := max(dCdx_Dec, dCdx_Inc);
      end
      else
        dCdx_Inc := 0;
    end;
  end;

end;

function TAdminEstados.deltaCosto_vxd_estrella(ird: integer; kpuntoT: integer;
  delta_xd: integer): NReal;
begin
  Result := constelacion.delta_costo_estrella_dx(kPuntoT, estrella_kr,
    estrella_kd, ird, delta_xd);
end;

function TAdminEstados.deltaCosto_vxd_continuo(ird: integer; kpuntoT: integer;
  delta_xd: integer): NReal;
begin
  if deterministico then
    if kPuntoT < nPuntosT then
      Result := ControladorDeterministico.GradInfos[kPuntoT - 1].Grad_Xr[ird]
    else
      Result := constelacion.delta_costo_continuo_dx(1, xr, xd, ird, delta_xd)
  else
    Result := constelacion.delta_costo_continuo_dx(kpuntoT, xr, xd, ird, delta_xd);
end;

function TAdminEstados.deltaCosto_vxr_continuo(irx: integer; kpuntoT: integer;
  delta_xr: NReal): NReal;
begin
  Result := constelacion.delta_costo_continuo_rx(kpuntoT, xr, xd, irx, delta_xr);
end;

function TAdminEstados.deltaCosto_vxr_continuo_DosEstados_UTE(irx: integer;
  irx2: integer; kpuntoT: integer; delta_xr: NReal; delta_xr2: NReal): NReal;

begin
  Result := constelacion.delta_costo_continuo_rx_DosEstados_UTE(kPuntoT,
    xr, xd, irx, irx2, delta_xr, delta_xr2);
end;



procedure TAdminEstados.SetEstadoToEstrella;
begin
  constelacion.cordenadasEstrella_icid(xr, xd, estrella_kr, estrella_kd);
end;


procedure TAdminEstados.SetEstadoToEstrella_indexada(
  vkr, vkd, idx_r, idx_d: TDAOfNInt);
begin
  constelacion.cordenadasEstrella_icid_indexada(xr, xd, vkr, vkd, idx_r, idx_d);
end;


constructor TAdminEstados.Create(_nVarsContinuas, _nVarsDiscretas, _nPuntosT: integer);
begin
  inherited Create;
  nVarsContinuas := _nVarsContinuas;
  nVarsDiscretas := _nVarsDiscretas;
  //1 porque el de la posicion 0 se tira y otro para almacenar el costo del
  //al final del ltimpo paso de tiempo.
  nPuntosT := _nPuntosT + 2;
  setLength(xr_def, nVarsContinuas);
  setLength(xd_def, nVarsDiscretas);
  setLength(xr, nVarsContinuas);
  setLength(xd, nVarsDiscretas);
  setLength(estrella_kr, nVarsContinuas);
  setLength(estrella_kd, nVarsDiscretas);
  constelacion := nil;
  deterministico := False;
  ControladorDeterministico := nil;
end;

procedure TAdminEstados.Free;
begin
  FechaIni.Free;
  fechaFin.Free;
  setLength(xr_def, 0);
  setLength(xd_def, 0);
  setLength(xr, 0);
  setLength(xd, 0);
  setLength(estrella_kr, 0);
  setLength(estrella_kd, 0);
  if constelacion <> nil then
    constelacion.Free;
  if ControladorDeterministico <> nil then
    ControladorDeterministico.Free;
  inherited Free;
end;

procedure TAdminEstados.Registrar_Continua(ixr: integer; xmin, xmax: NReal;
  nPuntos: integer; nombre: string; unidades: string);
begin
  if nPuntos < 2 then
    raise Exception.Create('TAdminEstados.Registrar_Continua, Nombre: ' +
      nombre + ' NPuntos :' + IntToStr(nPuntos) + '. Tiene que ser > 1');
  xr_def[ixr] := TDefVarContinua.Create(nombre, unidades, xmin, xmax, nPuntos);
end;

function TAdminEstados.indexOf_Continua(nombre: string): integer;
var
  k: integer;
  buscando: boolean;
begin
  k := 0;
  buscando := True;
  while buscando and (k < length(xr_def)) do
    if xr_def[k].nombre = nombre then
      buscando := False
    else
      Inc(k);
  if buscando then
    Result := -1
  else
    Result := k;
end;

procedure TAdminEstados.RegistrarTraduccion_Continua(ixr: integer;
  //xmin, xmax: NReal; nPuntos: integer;
  const nombre, unidades: string; funcTraduccion: TFuncTraduccion);
var
  i: integer;
begin
  xr_def[ixr].traduccion := TDefVarContinua.Create(nombre, unidades,
    0, 1, xr_def[ixr].nPuntos);
  for i := 0 to high(xr_def[ixr].x) do
    xr_def[ixr].traduccion.x[i] := funcTraduccion(xr_def[ixr].x[i]);
end;

procedure TAdminEstados.Registrar_Discreta(ixd: integer; npuntos: integer;
  const nombre, unidades: string);
begin
  xd_def[ixd] := TDefVarDiscreta.Create(nombre, unidades, Npuntos);
end;


function TAdminEstados.indexOf_Discreta(nombre: string): integer;
var
  k: integer;
  buscando: boolean;
begin
  k := 0;
  buscando := True;
  while buscando and (k < length(xd_def)) do
    if xd_def[k].nombre = nombre then
      buscando := False
    else
      Inc(k);
  if buscando then
    Result := -1
  else
    Result := k;
end;



{
 procedure TAdminEstados.RegsitrarTraduccion_Discreta(ixd: integer;
  // npuntos: integer;
  nombre: string; unidades: string);
 begin
  xd_def[ixd].traduccion := TDefVarDiscreta.Create(nombre, unidades, xd_def[ixd].Npuntos);
 end;
}

procedure TAdminEstados.CrearElEspacioTiempo(fechaIni, fechaFin: TFecha;
  horasDelPaso: NReal; const costoFuturo: TMatOfNReal; deterministico: boolean);
begin
  self.FechaIni := TFecha.Create_Clone(FechaIni);
  self.fechaFin := TFecha.Create_Clone(fechaFin);
  self.horasDelPaso := horasDelPaso;
  self.deterministico := deterministico;
  if not deterministico then
  begin
    Constelacion := TConstelacion.Create(xr_def, xd_def, nPuntosT, costoFuturo);
    ControladorDeterministico := nil;
  end
  else
  begin
    if costoFuturo <> nil then
      raise Exception.Create(
        'TAdminEstados.ClearElEspacioTiempo ... deterministico  and ( costofuturo <> nil ) no es posible'
        );
    // creo un solo frame de la constelacin para ser usado como FRAME FINAL
    // este frame debe ser llenado con el enganche que corresponda.
    Constelacion := TConstelacion.Create(xr_def, xd_def, 2, nil);
    ControladorDeterministico :=
      TControladorDeterministico.Create(length(xr_def), length(xd_def), nPuntosT - 1);
  end;
end;

procedure AlInicio;
begin
  registrarClaseDeCosa(TEngancheCFReal.ClassName, TEngancheCFReal);
  registrarClaseDeCosa(TEngancheCFEntero.ClassName, TEngancheCFEntero);
end;

procedure AlFinal;
begin

end;

end.
